/* real time linux task for reading samples from the MAX192 chip       */
/* MAX192 is a 12 bit, 8 channel multiplexed, serial AD-converter chip */

#define PORT SER

#include <linux/errno.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <rtl.h>
#include <time.h>
#include <asm/io.h>

#include <pthread.h>

#include <rtl_sched.h>
#include <rtl_fifo.h>
#include "control.h"

pthread_t mytask;
int       sr;
char      channels;

/* read channels from AD-card in unipolar mode */
/* voltage range is 0 V to +4.0 V              */

int kanal0(int port) {
  int  vout=0;
  int  i, addit;

  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);                /* pull SCLK low             */
    outb((0xFD & inb(port)), port);                /* pull DIN  low, do nothing */
    outb((0x01 | inb(port)), port);                /* pull SCLK high            */
    if (((inb(port+2) & 0x10))==0x10) vout+=addit; /* get bits 11-5             */
    addit=addit/2;
  }
  outb((0xFE & inb(port)), port);                /* pull SCLK low             */
  outb((0x02 | inb(port)), port);                /* set start bit in DIN to 1 */
  outb((0x01 | inb(port)), port);                /* pull SCLK high            */
  if (((inb(port+2) & 0x10))==0x10) vout+=addit; /* get bit 4                 */
  addit=addit/2;
  for (i=0; i<3; i++) {
    outb((0xFE & inb(port)), port); /* pull SCLK low                             */
    outb((0xFD & inb(port)), port); /* set DIN low, select channel bits 2-4 to 0 */
    outb((0x01 | inb(port)), port); /* pull SCLK high                            */
    if (((inb(port+2) & 0x10))==0x10) vout+=addit; /* get bits 3-1               */
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port); /* pull SCLK low            */
  outb((0x02 | inb(port)), port); /* set DIN high, unipolar   */
  outb((0x01 | inb(port)), port); /* pull SCLK high           */
  if (((inb(port+2) & 0x10))==0x10) vout+=addit; /* get bit 0 */
  addit=addit/2;

  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal1(int port) {
  int vout=0;
  int i, addit;

  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }
  for (i=0; i<2; i++) {
    outb((0xFE & inb(port)), port);
    outb((0x02 | inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }
  for (i=0; i<2; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal2(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal3(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal4(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal5(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  for (i=0; i<3; i++) {
    outb((0xFE & inb(port)), port);
    outb((0x02 | inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal6(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  outb((0xFE & inb(port)), port);
  outb((0xFD & inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  for (i=0; i<2; i++) {
    outb((0xFE & inb(port)), port);
    outb((0x02 | inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;

  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

int kanal7(int port) {
  int vout=0;
  int i, addit;

  /* get bits */
  addit=2048;
  for (i=0; i<7; i++) {
    outb((0xFE & inb(port)), port);
    outb((0xFD & inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }

  for (i=0; i<4; i++) {
    outb((0xFE & inb(port)), port);
    outb((0x02 | inb(port)), port);
    outb((0x01 | inb(port)), port);
    if (((inb(port+2) & 0x10))==0x10) vout+=addit;
    addit=addit/2;
  }
  
  outb((0xFE & inb(port)), port);
  outb((0x02 | inb(port)), port);
  outb((0x01 | inb(port)), port);
  if (((inb(port+2) & 0x10))==0x10) vout+=addit;
  addit=addit/2;
  
  outb(0xFE & inb(port), port); /* pull SCLK low              */
  outb(0x02 | inb(port), port); /* set DIN high, single ended */
  outb(0x01 | inb(port), port); /* pull SCLK high             */

  outb(0xFE & inb(port), port); /* pull SCLK low                           */
  outb(0x02 | inb(port), port); /* set DIN high, PD1=1 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                          */

  outb(0xFE & inb(port), port); /* pull SCLK low                          */
  outb(0xFD & inb(port), port); /* set DIN low, PD0=0 internal clock mode */
  outb(0x01 | inb(port), port); /* pull SCLK high                         */

  outb(0xFE & inb(port), port); /* pull SCLK low                               */
  outb(0xFD & inb(port), port); /* set DIN low, empty bit 16 of shift register */ 
  outb(0x01 | inb(port), port); /* pull SCLK high                              */

  return(vout);
}

/* .................................. */
/* function to write samples to fifos */
/* .................................. */
void *fun(void *t) {
  int data, prech;

  int kanal0(int port);
  int kanal1(int port);
  int kanal2(int port);
  int kanal3(int port);
  int kanal4(int port);
  int kanal5(int port);
  int kanal6(int port);
  int kanal7(int port);

  prech=-1;
  while(1) {
    /* FIFO's 0-7 transfer the samples from channel 1 to 8                       */
    /* FIFO 9 is an indicator whether a FIFO is full when trying to write to it */
    if (channels==0) pthread_wait_np();
    if ((channels & 1)==1) {
      data=kanal0(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=0;
      pthread_wait_np();
    }
    if ((channels & 2)==2) {
      data=kanal1(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=1;
      pthread_wait_np();
    }
    if ((channels & 4)==4) {
      data=kanal2(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=2;
      pthread_wait_np();
    }
    if ((channels & 8)==8) {
      data=kanal3(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=3;
      pthread_wait_np();
    }
    if ((channels & 16)==16) {
      data=kanal4(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=4;
      pthread_wait_np();
    }
    if ((channels & 32)==32) {
      data=kanal5(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=5;
      pthread_wait_np();
    }
    if ((channels & 64)==64) {
      data=kanal6(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=6;
      pthread_wait_np();
    }
    if ((channels & 128)==128) {
      data=kanal7(PORT);
      if (prech>=0)
        if (rtf_put(prech, (char *) &data, 2) < 0)
	  rtf_put(9, (char *) &prech, sizeof(prech));
      prech=7;
      pthread_wait_np();
    }
  }
}

int my_handler(unsigned int fifo) {
  struct my_msg_struct msg;
  int err;

  /* FIFO 8 transfers commands to control the driver */
  while ((err=rtf_get(8, &msg, sizeof(msg))) == sizeof(msg)) {
    switch(msg.command) {
    case START_TASK:
      pthread_make_periodic_np(mytask, gethrtime(), HRTICKS_PER_SEC*(float)(1.0/(float)sr));
      pthread_wakeup_np(mytask);
      break;
    case STOP_TASK:
      pthread_suspend_np(mytask);
      break;
    case SET_CH:
      channels=(unsigned char)msg.value;
      break;
    case SET_POL:
      /* do nothing, MAX192 does not support bipolar mode */
      break;
    case SET_SR:
      sr=msg.value;
      break;
    default:
      return -EINVAL;
    }
  }
  if (err!=0) {
    return -EINVAL;
  }
  return 0;
}

int init_module(void) {
  pthread_attr_t     attr;
  struct sched_param sched_param;
  int inp;

  sr=1;
  channels=0x00;
  outb(0x40, PORT-1);                    /* pull TXD high for power supply */
  inp=inb(PORT); outb(0x01 | inp, PORT); /* pull DTR high for power supply */
  rtf_create(0,10000);
  rtf_create(1,10000);
  rtf_create(2,10000);
  rtf_create(3,10000);
  rtf_create(4,10000);
  rtf_create(5,10000);
  rtf_create(6,10000);
  rtf_create(7,10000);
  rtf_create(8,100);
  rtf_create(9,100);

  pthread_attr_init(&attr);
  sched_param.sched_priority = 2;
  pthread_attr_setschedparam(&attr, &sched_param);
  pthread_create(&mytask,  &attr, fun, (void *)1);
  pthread_setfp_np(mytask,1);
  rtf_create_handler(8, &my_handler);
  return 0;
}

void cleanup_module(void)
{
  rtf_destroy(0);
  rtf_destroy(1);
  rtf_destroy(2);
  rtf_destroy(3);
  rtf_destroy(4);
  rtf_destroy(5);
  rtf_destroy(6);
  rtf_destroy(7);
  rtf_destroy(8);
  rtf_destroy(9);
  pthread_delete_np(mytask);
}
